#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "slpd.h"
#include "slp_message.h"
#include "slp_property.h"
#include "slp_network.h"
#include "slpd_database.h"
#include "slpd_regfile.h"
#include "slpd_property.h"
#include "slpd_log.h"


#define SLP_NETWORK_TIMED_OUT		-19
#define SLP_MEMORY_ALLOC_FAILED		-21
#define SLP_NETWORK_ERROR		-23

static int SLPDUnicastRqstRply(int sock, struct sockaddr_storage * destaddr, struct sockaddr_storage * localaddr,
    const char * langtag, char * buf, char buftype, int bufsize,
    int (*callback)(SLPMessage * message, void *cookie), void * cookie)
{
    struct timeval timeout;
    SLPBuffer sendbuf = 0;
    SLPBuffer recvbuf = 0;
    SLPMessage * message = 0;
    int result = 0;
    int langtaglen = 0; 
    int xid = 0;
    int mtu = 0;
    int size = 0;
    int timeouts[1];

    /* Save off a few things we don't want to recalculate */
    langtaglen = strlen(langtag);
    xid = SLPXidGenerate();
    mtu = SLPPropertyAsInteger("net.slp.MTU");
    sendbuf = SLPBufferAlloc(mtu);
    if(sendbuf == 0)
    {
        result = SLP_MEMORY_ALLOC_FAILED;
	goto FINISHED;
    }
    SLPPropertyAsIntegerVector("net.slp.unicastTimeouts", timeouts, 1);
    timeout.tv_sec = timeouts[0] / 1000;
    timeout.tv_usec = (timeouts[0] % 1000) * 1000;
    size = 16 + langtaglen + bufsize;
    if((sendbuf = SLPBufferRealloc(sendbuf, size)) == 0)
    {
	result = SLP_MEMORY_ALLOC_FAILED;
	goto FINISHED;
    }
    sendbuf->curpos = sendbuf->start;

    /* Add the header to the send buffer */
    /*version*/
    *sendbuf->curpos++ = 2;
    /*function id*/
    *sendbuf->curpos++ = buftype;
    /*length*/
    PutUINT24(&sendbuf->curpos, size);
    /*flags*/
    PutUINT16(&sendbuf->curpos, SLP_FLAG_UCAST);  /*this is a unicast */
    /*ext offset*/
    PutUINT24(&sendbuf->curpos, 0);
    /*xid*/
    PutUINT16(&sendbuf->curpos, xid);
    /*lang tag len*/
    PutUINT16(&sendbuf->curpos, langtaglen);
    /*lang tag*/
    memcpy(sendbuf->curpos, langtag, langtaglen);
    sendbuf->curpos += langtaglen;
    /*prlist*/
    PutUINT16(&sendbuf->curpos, 0);

    /* Add the rest of the message */
    memcpy(sendbuf->curpos, buf, bufsize);
    sendbuf->curpos += bufsize;

    /* send the send buffer */
    result = SLPNetworkSendMessage(sock, SOCK_STREAM, sendbuf, sendbuf->curpos - sendbuf->start, destaddr, &timeout);
    if (result != 0)
    {
	result = errno == ETIMEDOUT ? SLP_NETWORK_TIMED_OUT : SLP_NETWORK_ERROR;
	goto FINISHED;
    }
    result = SLPNetworkRecvMessage(sock, SOCK_STREAM, &recvbuf, destaddr, &timeout);
    if (result != 0)
    {
	result = errno == ETIMEDOUT ? SLP_NETWORK_TIMED_OUT : SLP_NETWORK_ERROR;
	goto FINISHED;
    }
    if(AS_UINT16(recvbuf->start + 10) != xid)
    {
	result = SLP_NETWORK_ERROR;
	goto FINISHED;
    }
    message = SLPMessageAlloc();
    result = SLPMessageParseBuffer(destaddr, localaddr, recvbuf, message);
    if (result == 0)
    {
	result = callback(message, cookie);
    }
FINISHED:
    SLPMessageFree(message);
    SLPBufferFree(sendbuf);
    SLPBufferFree(recvbuf);
    return result;
}

typedef struct SLPURL {
    struct SLPURL *next;
    struct SLPURL *last;
    char *serviceURL;
    char *attrs;
    char* scopelist;
    char* serviceType;      
    unsigned short ltime;
} SLPUrl;


/* Cache collection structure */
typedef struct SLPUrlList {
    char * services; /* list of all services */
    SLPUrl * slpUrl; /* linked list of URLs for all services */
    SLPUrl * currentSLPUrl;  /* next location to be used for update */
    char * currentScope;
    char * currentServiceType;
} SLPUrlList;  

static SLPUrl * AllocateSLPUrl()
{
    SLPUrl* slpUrl          = (SLPUrl*)malloc(sizeof(SLPUrl));
    slpUrl->serviceURL      = NULL;
    slpUrl->ltime           = 0;
    slpUrl->scopelist       = NULL;
    slpUrl->serviceType     = NULL;
    slpUrl->attrs           = NULL;
    slpUrl->next            = NULL;
    slpUrl->last            = NULL;
    return slpUrl;
}


static void CleanUpSLPUrlList(SLPUrlList* slpUrlList)
{
    SLPUrl* slpUrl,*slpUrlNext;

    slpUrl = slpUrlList->slpUrl;
    while(slpUrl)
    {
	if (slpUrl->serviceURL)
	    free(slpUrl->serviceURL);
	if (slpUrl->scopelist)
	    free(slpUrl->scopelist);
	if (slpUrl->serviceType)
	    free(slpUrl->serviceType);
	if (slpUrl->attrs)
	    free(slpUrl->attrs);
	slpUrlNext = slpUrl->next;
	free(slpUrl);
	slpUrl = slpUrlNext;
    }
    slpUrlList->slpUrl = NULL;
    if (slpUrlList->currentScope != NULL)
	free(slpUrlList->currentScope);
    if(slpUrlList->services != NULL)
	free(slpUrlList->services);
    if(slpUrlList->currentServiceType != NULL)
	free(slpUrlList->currentServiceType);
    free(slpUrlList);
}

static int SLPSrvTCallBack(SLPMessage * message, void * cookie)
{
    if (message->header.functionid != SLP_FUNCT_SRVTYPERPLY)
	return SLP_NETWORK_ERROR;
    if (message->body.srvtyperply.errorcode != 0)
	return message->body.srvtyperply.errorcode;
    /* null terminate as in libslp */
    ((char *)message->body.srvtyperply.srvtypelist)[message->body.srvtyperply.srvtypelistlen] = 0;
    ((SLPUrlList*)cookie)->services = strdup(message->body.srvtyperply.srvtypelist);
    return 0;
}

static int SLPSrvCallBack(SLPMessage * message, void * cookie)
{
    SLPUrl *slpUrl = NULL;
    char scopelist[4096];
    int i;
    SLPUrlEntry *srvurl;

    if (message->header.functionid != SLP_FUNCT_SRVRPLY)
	return SLP_NETWORK_ERROR;
    if (message->body.srvrply.errorcode != 0)
	return message->body.srvrply.errorcode;
    for (i=0; i<message->body.srvrply.urlcount; i++)
    {
	srvurl = message->body.srvrply.urlarray + i;
	/* null terminate url as in libslp, overwrites authcount */
	((char *)srvurl->url)[srvurl->urllen] = 0;
        for (slpUrl = ((SLPUrlList*)cookie)->slpUrl; slpUrl ; slpUrl = slpUrl->next)
        {
	    /* Check whether the same service URL is available as part of different scope*/
	    if( (slpUrl->serviceURL != NULL) && ( strcasecmp( slpUrl->serviceURL,srvurl->url) == 0))
		break;
	}
	if (slpUrl)
	{
	    snprintf(scopelist,sizeof(scopelist),"%s,%s",slpUrl->scopelist,((SLPUrlList*)cookie)->currentScope);
	    free(slpUrl->scopelist);
	    slpUrl->scopelist = strdup(scopelist);
	}
	else
	{
	    slpUrl                  = AllocateSLPUrl();
	    slpUrl->serviceURL      = strdup(srvurl->url);
	    slpUrl->ltime           = srvurl->lifetime;
	    slpUrl->scopelist       = strdup(((SLPUrlList*)cookie)->currentScope);
	    slpUrl->serviceType     = strdup(((SLPUrlList*)cookie)->currentServiceType);
	    slpUrl->attrs           = NULL;
	    slpUrl->next            = ((SLPUrlList*)cookie)->slpUrl;
	    if(((SLPUrlList*)cookie)->slpUrl)
		((SLPUrlList*)cookie)->slpUrl->last = slpUrl;
	    ((SLPUrlList*)cookie)->slpUrl= slpUrl;
        }
    }
    return 0;
}

static int SLPSrvAttrCallBack(SLPMessage * message, void * cookie)
{
    SLPUrl *lslpUrl = ((SLPUrlList*)cookie)->currentSLPUrl;
    if (message->header.functionid != SLP_FUNCT_ATTRRPLY)
	return SLP_NETWORK_ERROR;
    if (message->body.attrrply.errorcode != 0)
	return message->body.attrrply.errorcode;

    /* null terminate as in libslp */
    ((char *)message->body.attrrply.attrlist)[message->body.attrrply.attrlistlen] = 0;
    lslpUrl->attrs = strdup(message->body.attrrply.attrlist);
    return 0;
}

static char * createreq(int * sizep, char * url, char * scope, char * predicate, char * spi)
{
  char *buf, *cur;
  int urllen = url ? strlen(url) : 0;
  int scopelen = scope ? strlen(scope) : 0;
  int predicatelen = predicate ? strlen(predicate) : 0;
  int spilen = spi ? strlen(spi) : 0;
  buf = malloc(2 + urllen + 2 + scopelen + 2 + predicatelen + 2 + spilen);
  cur = buf;
  if (url)
  {
	TO_UINT16(cur, urllen);
	cur += 2;
	memcpy(cur, url, urllen);
	cur += urllen;
  }
  if (scope)
  {
	TO_UINT16(cur, scopelen);
	cur += 2;
	memcpy(cur, scope, scopelen);
	cur += scopelen;
  }
  if (predicate)
  {
	TO_UINT16(cur, predicatelen);
	cur += 2;
	memcpy(cur, predicate, predicatelen);
	cur += predicatelen;
  }
  if (spi)
  {
	TO_UINT16(cur, spilen);
	cur += 2;
	memcpy(cur, spi, spilen);
	cur += spilen;
  }
  *sizep = cur - buf;
  return buf;
}


int getSLPServiceURLs(int sock, struct sockaddr_storage * destaddr, struct sockaddr_storage * localaddr)
{
    char *strng, *services;
    int gresult = 0, result;
    SLPUrl * slpUrl;
    char *scope = NULL, *scopelist = NULL;
    char scopeptr[4096],serviceptr[4096];
    SLPMessage * msg;
    SLPBuffer buf;
    const char *langtag;
    char *outbuf;
    int bufsize;
    char *srvtype, *srvtype_end;

    SLPUrlList* slpUrlList  = (SLPUrlList*)malloc(sizeof(SLPUrlList));
    slpUrlList->slpUrl      = NULL;
    slpUrlList->services    = NULL;
    slpUrlList->currentServiceType  = NULL;
    slpUrlList->currentScope  = NULL;

    langtag = SLPPropertyGet("net.slp.locale", 0, 0);

    scopelist = strdup(G_SlpdProperty.useScopes);

    for(scope = strtok_r(scopelist,",",(char**)&scopeptr); scope ; scope = strtok_r(NULL,",",(char**)&scopeptr))
    {
	slpUrlList->currentScope = strdup(scope);
	outbuf = createreq(&bufsize, "", scope, NULL, NULL);
	TO_UINT16(outbuf, 0xffff); /* 0xffff indicates all service types */
	result = SLPDUnicastRqstRply(sock, destaddr, localaddr, langtag, outbuf, SLP_FUNCT_SRVTYPERQST, bufsize, SLPSrvTCallBack, slpUrlList);
	free(outbuf);
	if (result)
	{
	    gresult = result;	/* remember error with that scope */
	    SLPDLog("Error: SLPFindSrvTypes %d\n",result);
	    continue;
	}
	if(slpUrlList->services)
	{
	    services = strdup(slpUrlList->services);
	    for(strng = strtok_r(services, ",",(char**)&serviceptr); strng ; strng = strtok_r(NULL, ",",(char**)&serviceptr))
	    {
		if (!strcasecmp(strng, SLP_SA_SERVICE_TYPE))
		    continue;
		if (!strcasecmp(strng, SLP_DA_SERVICE_TYPE))
		    continue;
		slpUrlList->currentServiceType = strdup(strng);
		outbuf = createreq(&bufsize, strng, scope, "", "");
		result = SLPDUnicastRqstRply(sock, destaddr, localaddr, langtag, outbuf, SLP_FUNCT_SRVRQST, bufsize, SLPSrvCallBack, slpUrlList);
		free(outbuf);
		free(slpUrlList->currentServiceType);
		slpUrlList->currentServiceType = NULL;
		if(result != 0)
		{
		    SLPDLog("Error: SLPFindSrvs %d\n", result);
		    continue;
		}
	    }
	    free(services);
	}
	if (slpUrlList->currentScope != NULL)
	{
	    free(slpUrlList->currentScope);
	    slpUrlList->currentScope = NULL;
	}
	if(slpUrlList->services != NULL)
	{
	    free(slpUrlList->services);
	    slpUrlList->services = NULL;
	}
    }

    /* we now have collected all services, fetch the attributes */

    for(slpUrl = slpUrlList->slpUrl; slpUrl ; slpUrl = slpUrl->next)
    {
	slpUrl->attrs = NULL;
	slpUrlList->currentSLPUrl = slpUrl;
	outbuf = createreq(&bufsize, slpUrl->serviceURL, slpUrl->scopelist, "", "");
	result = SLPDUnicastRqstRply(sock, destaddr, localaddr, langtag, outbuf, SLP_FUNCT_ATTRRQST, bufsize, SLPSrvAttrCallBack, slpUrlList);
	free(outbuf);
	if(result != 0)
	{
	    SLPDLog("Error: SLPFindAttrs %d\n", result);
	    continue;
	}
	srvtype = strdup(slpUrl->serviceURL);
	srvtype_end = strstr(srvtype, "://");
	if (srvtype_end)
	    *srvtype_end = 0;
	if (SLPDCreateSrvReg(SLP_REG_SOURCE_PULL_PEER_DA,
                             strlen(slpUrl->serviceURL), slpUrl->serviceURL,
                             strlen(langtag), (char *)langtag,
                             strlen(srvtype), srvtype,
                             strlen(slpUrl->scopelist), slpUrl->scopelist,
                             slpUrl->attrs ? strlen(slpUrl->attrs) : 0, slpUrl->attrs,
                             slpUrl->ltime, &msg, &buf) == 0)
	{
	    msg->peer = *destaddr;
	    SLPDDatabaseReg(msg, buf);
	}
	free(srvtype);
    }

    CleanUpSLPUrlList(slpUrlList);
    if(scopelist != NULL)
	free(scopelist);
    return gresult;
}

