/*-------------------------------------------------------------------------
 * Copyright (C) 2000 Caldera Systems, Inc
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *    Neither the name of Caldera Systems nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * `AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE CALDERA
 * SYSTEMS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *-------------------------------------------------------------------------*/

/** Main entry point for the slpd process.
 *
 * @file       slpd_main.c
 * @author     Matthew Peterson, John Calcote (jcalcote@novell.com)
 * @attention  Please submit patches to http://www.openslp.org
 * @ingroup    SlpdCode
 */

#include "slp_types.h"

#include "slpd_log.h"
#include "slpd_socket.h"
#include "slpd_incoming.h"
#include "slpd_outgoing.h"
#include "slpd_database.h"
#include "slpd_cmdline.h"
#include "slpd_knownda.h"
#include "slpd_property.h"
#include "slpd.h"

#ifdef ENABLE_SLPv2_SECURITY
# include "slpd_spi.h"
#endif

#include "slp_xmalloc.h"
#include "slp_xid.h"
#include "slp_net.h"
#include "slp_network.h"

#include <systemd/sd-daemon.h>

int G_SIGALRM;
int G_SIGTERM;
int G_SIGHUP;
#ifdef DEBUG
int G_SIGINT;     /* Signal being used for dumping registrations */
int G_SIGUSR1;    /* Signal being used to dump information about the database */
#endif
#include "slpd_initda.h"

char *reg_file_dir = "/etc/slp.reg.d";
FILE *DABackupfp;

static void SLPDOpenDABackupFile()
{
   FILE *fp;
   char filename[1024];
   snprintf(filename, sizeof(filename), "%s/slpd/%s", reg_file_dir, "DABackup");

   fp = fopen(filename, "a+");
   if (!DABackupfp && !fp)
      SLPDLog("Could not open DABackup file\n");
   if (fp)
   {
      if (DABackupfp)
         fclose(DABackupfp);
      DABackupfp = fp;
   }
}

/** Configures fd_set objects with sockets.
 *
 * @param[in] socklist - The list of sockets that is being currently
 *    monitored by OpenSLP components.
 * @param[out] highfd - The address of storage for returning the value
 *    of the highest file descriptor (number) in use.
 * @param[out] fdset - The fdset to fill with read/write descriptors.
 */
void LoadFdSets(SLPList * socklist, SLPD_fdset * fdset)
{
#if HAVE_POLL

   SLPDSocket * sock = 0;
   SLPDSocket * del = 0;

   sock = (SLPDSocket *)socklist->head;
   while (sock)
   {
      if (fdset->used == fdset->allocated)
      {
         fdset->allocated += 32;
         if (fdset->used)
            fdset->fds = xrealloc(fdset->fds, fdset->allocated * sizeof(*fdset->fds));
         else
            fdset->fds = xmalloc(fdset->allocated * sizeof(*fdset->fds));
         if (!fdset->fds)
            SLPDFatal("No memory for fdset.\n");
      }
      fdset->fds[fdset->used].fd = sock->fd;
      fdset->fds[fdset->used].events = 0;
      fdset->fds[fdset->used].revents = 0;
      switch (sock->state)
      {
         case DATAGRAM_UNICAST:
         case DATAGRAM_MULTICAST:
         case DATAGRAM_BROADCAST:
            fdset->fds[fdset->used].events |= POLLIN;
            break;

         case SOCKET_LISTEN:
            if (socklist->count < SLPD_MAX_SOCKETS)
               fdset->fds[fdset->used].events |= POLLIN;
            break;

         case STREAM_READ:
         case STREAM_READ_FIRST:
            fdset->fds[fdset->used].events |= POLLIN;
            break;

         case STREAM_WRITE:
         case STREAM_WRITE_FIRST:
         case STREAM_CONNECT_BLOCK:
            fdset->fds[fdset->used].events |= POLLOUT;
            break;

         case SOCKET_CLOSE:
            del = sock;
            break;

         default:
            break;
      }

      if (fdset->fds[fdset->used].events)
         sock->fdsetnr = fdset->used++;
      else
         sock->fdsetnr = -1;

      sock = (SLPDSocket*)sock->listitem.next;

      if (del)
      {
         SLPDSocketFree((SLPDSocket*)SLPListUnlink(socklist,(SLPListItem*)del));
         del = 0;
      }
   }

#else   /* HAVE_POLL */

   SLPDSocket* sock = 0;
   SLPDSocket* del = 0;

   sock = (SLPDSocket*)socklist->head;
   while(sock)
   {
      if (sock->fd > fdset->highfd)
         fdset->highfd = sock->fd;

      switch(sock->state)
      {
         case DATAGRAM_UNICAST:
         case DATAGRAM_MULTICAST:
         case DATAGRAM_BROADCAST:
            FD_SET(sock->fd, &fdset->readfds);
            break;

         case SOCKET_LISTEN:
            if (socklist->count < SLPD_MAX_SOCKETS)
               FD_SET(sock->fd, &fdset->readfds);
            break;

         case STREAM_READ:
         case STREAM_READ_FIRST:
            FD_SET(sock->fd, &fdset->readfds);
            break;

         case STREAM_WRITE:
         case STREAM_WRITE_FIRST:
         case STREAM_CONNECT_BLOCK:
            FD_SET(sock->fd, &fdset->writefds);
            break;

         case SOCKET_CLOSE:
            del = sock;
            break;

         default:
            break;
      }
      sock = (SLPDSocket*)sock->listitem.next;
      if (del)
      {
         SLPDSocketFree((SLPDSocket *)SLPListUnlink(socklist,
               (SLPListItem*)del));
         del = 0;
      }
   }

#endif  /* HAVE_POLL */
}

/** Handles a SIG_TERM signal from the system.
 */
void HandleSigTerm(void)
{
   struct timeval timeout;
   SLPD_fdset fdset;
   int fdcount = 0;

   SLPD_fdset_init(&fdset);

   SLPDLog("****************************************\n");
   SLPDLogTime();
   SLPDLog("SLPD daemon shutting down\n");
   SLPDLog("****************************************\n");

   /* write backup file if configured */
   if (G_SlpdProperty.isDA && G_SlpdProperty.isDABackup && DABackupfp)
      SLPDDatabaseWriteDABackup(DABackupfp);

   /* unregister with all DAs */
   SLPDKnownDADeinit();

   timeout.tv_sec  = 5;
   timeout.tv_usec = 0;

   /* Do a dead DA passive advert to tell everyone we're goin' down */
   SLPDKnownDAPassiveDAAdvert(0, 1);

   /* close all incoming sockets */
   SLPDIncomingDeinit();

   /* if possible wait until all outgoing socket are done and closed */
   while (SLPDOutgoingDeinit(1))
   {
      SLPD_fdset_reset(&fdset);
      LoadFdSets(&G_OutgoingSocketList, &fdset);
#if HAVE_POLL
      fdcount = poll(fdset.fds, fdset.used, timeout.tv_sec * 1000 + timeout.tv_usec / 1000);
#else
      fdcount = select((int)fdset.highfd + 1, &fdset.readfds, &fdset.writefds, 0, &timeout);
#endif
      if (fdcount == 0)
         break;
      SLPDOutgoingHandler(&fdcount, &fdset);
   }

   SLPDOutgoingDeinit(0);

   SLPDLog("****************************************\n");
   SLPDLogTime();
   SLPDLog("SLPD daemon shut down\n");
   SLPDLog("****************************************\n");

#ifdef DEBUG
# ifdef ENABLE_SLPv2_SECURITY
   SLPDSpiDeinit();
# endif
   SLPDDatabaseDeinit();
   SLPDPropertyDeinit();
   SLPDLogFileClose();
   xmalloc_deinit();
#endif
}

#ifdef DEBUG
/** Handles a SIGUSR1 signal from the system.
 *
 * @internal
 */
static void HandleSigUsr1(void)
{
   /* Output some debug data */
   extern void SLPDDatabaseUsr1(void);
   SLPDDatabaseUsr1();
}
#endif

/** Handles a SIG_HUP signal from the system.
 */
void HandleSigHup(void)
{
   /* Reinitialize */
   SLPDLog("****************************************\n");
   SLPDLogTime();
   SLPDLog("SLPD daemon reset by SIGHUP\n");
   SLPDLog("****************************************\n\n");

   /* write backup file if configured */
   if (G_SlpdProperty.isDA && G_SlpdProperty.isDABackup && DABackupfp)
      SLPDDatabaseWriteDABackup(DABackupfp);

   /* unregister with all DAs */
   SLPDKnownDADeinit();

   /* re-read properties */
   SLPDPropertyReinit();

#ifdef ENABLE_SLPv2_SECURITY
   /* Re-initialize SPI stuff*/
#if 0	/* does not work in chroot, sorry */
   SLPDSpiInit(G_SlpdCommandLine.spifile);
#endif
#endif

   /* Re-read the static registration file (slp.reg)*/
   SLPDDatabaseReInit();

   /* Re-read the backup file if configured */
   if (G_SlpdProperty.isDA && G_SlpdProperty.isDABackup && DABackupfp)
      SLPDDatabaseReadDABackup(DABackupfp);

   /* Reopen listening sockets */
   SLPDIncomingReinit();

   /* Rebuild Known DA database */
   SLPDKnownDAInit();

   SLPDLog("****************************************\n");
   SLPDLogTime();
   SLPDLog("SLPD daemon reset finished\n");
   SLPDLog("****************************************\n\n");
   SLPDLog("Agent Interfaces = %s\n", G_SlpdProperty.interfaces);
}

/** Handles a SIG_ALRM signal from the system.
 */
void HandleSigAlrm(void)
{
   SLPDIncomingAge(SLPD_AGE_INTERVAL);
   SLPDOutgoingAge(SLPD_AGE_INTERVAL);
   SLPDKnownDAImmortalRefresh(SLPD_AGE_INTERVAL);
   SLPDKnownDAPassiveDAAdvert(SLPD_AGE_INTERVAL, 0);
   SLPDKnownDAStaleDACheck(SLPD_AGE_INTERVAL);
   SLPDKnownDAActiveDiscovery(SLPD_AGE_INTERVAL);
   SLPDDatabaseAge(SLPD_AGE_INTERVAL, G_SlpdProperty.isDA);
   SLPDDatabaseWatcher();
}

#ifdef DEBUG
/** Handles a SIG_INT signal from the system.
 *
 * @internal
 */
static void HandleSigInt(void)
{
   SLPDIncomingSocketDump();
   SLPDOutgoingSocketDump();
   SLPDKnownDADump();
   SLPDDatabaseDump();
}
#endif

#ifndef _WIN32
/** Check a pid file to see if slpd is already running.
 *
 * The pid file contains the PID of the process that is already running.
 *
 * @param[in] pidfile - The name of a file to read.
 *
 * @return Zero on success or a non-zero value on failure.
 *
 * @internal
 */
static int CheckPid(const char * pidfile)
{
   pid_t pid;
   FILE * fd;
   char pidstr[14];

   /* make sure that we're not running already
      read the pid from the file
    */
   fd = fopen(pidfile, "r");
   if (fd)
   {
      memset(pidstr,0,14);
      (void)fread(pidstr,13,1,fd);
      pid = atoi(pidstr);
      if (pid && kill(pid, 0) == 0)
         return -1;  /* we are already running */
      fclose(fd);
   }
   return 0;
}

/** Write the pid file.
 *
 * @param[in] pidfile - The name of the file to write.
 * @param[in] pid - The PID value to write to @p pidfile.
 *
 * @return Zero on success, or a non-zero value on failure.
 *
 * @internal
 */
/* static */ int WritePid(const char* pidfile, pid_t pid)
{
   FILE * fd;
   char pidstr[14];

   /* write my pid to the pidfile */
   fd = fopen(pidfile, "w");
   if(fd)
   {
      sprintf(pidstr, "%i", (int)pid);
      fwrite(pidstr, strlen(pidstr), 1, fd);
      fclose(fd);
   }
   return 0;
}

/** Daemonize the calling process.
 *
 * Turn the calling process into a daemon (detach from tty setuid(), etc.
 *
 * @param[in] pidfile - The name of a file to which the process id should
 *    be written.
 *
 * @return Zero on success, or a non-zero value if slpd could not daemonize
 *    (or if slpd is already running).
 *
 * @internal
 */
static int Daemonize(const char * pidfile)
{
   pid_t pid;

   /* fork() if we should detach */
   if (G_SlpdCommandLine.detach)
      pid = fork();
   else
      pid = getpid();

   /* parent or child? */
   switch(pid)
   {
      case -1:
         return -1;

      case 0:
         /* child lives */
         break;

      default:
      {
         /* parent writes pid (or child) pid file and dies */
         FILE * fd = fopen(pidfile,"w");
         if (fd)
         {
            char pidstr[14];
            sprintf(pidstr,"%i",(int)pid);
            fwrite(pidstr,strlen(pidstr),1,fd);
            fclose(fd);
         }
         if (G_SlpdCommandLine.detach)
            exit(0);
         break;
      }
   }

   /* Change directory, close files, reset umask, detach from
      console, and (maybe) drop permissions if daemonizing (-d) */

   if (G_SlpdCommandLine.detach)
   {
      int i;

      /* change directory to root */
      (void)chdir("/");

      /* close all open file handles */
      for (i = 0; i < 8192; i++)
         close(i);

      umask(0); /* change file file mode mask */
      setsid(); /* create a new sid for the child process */

   }
   return 0;
}

/** Drop privileges to reduce security risk.
 *
 * @return Zero on success, or a non-zero value if slpd could not daemonize
 *    (or if slpd is already running).
 *
 * @internal
 */
static int DropPrivileges()
{
#ifndef SETUIDS
   /* suid to daemon */

   /* TODO: why do the following lines mess up my signal handlers? */

   /* IPv6 Operation requires that the process owner has permissons to
      open multicast sockets under IPv6, this process owner is 'daemon'.
      As a quick workaround for correct IPv6 operation, to run this
      process as root, define SETUIDS. */

   /* TODO: warn if 'daemon' user has insufficient privileges and ipv6 requested.*/
   /* TODO: allow different user to be specified as process owner. */

   struct passwd * pwent = getpwnam("openslp");
   if (pwent)
   {
      if (chroot(reg_file_dir))
         return 1;
      reg_file_dir = ".";

      if (setgroups(1, &pwent->pw_gid) < 0 || setgid(pwent->pw_gid) < 0
            || setuid(pwent->pw_uid) < 0)
      {
         /* TODO: should we log here and return fail */
         return 1;
      }
   } else {
       return 1;
   }
#endif
   return 0;
}

/** Handles all registered signals from the system.
 *
 * @param[in] signum - The signal number to handle.
 *
 * @internal
 */
static void SignalHandler(int signum)
{
   switch(signum)
   {
      case SIGALRM:
         G_SIGALRM = 1;
         break;

      case SIGTERM:
         G_SIGTERM = 1;
         break;

      case SIGHUP:
         G_SIGHUP = 1;
         break;

#ifdef DEBUG
      case SIGINT:
         G_SIGINT = 1;
         break;

      case SIGUSR1:
         G_SIGUSR1 = 1;
         break;
#endif

      case SIGPIPE:
      default:
         break;
   }
}

/** Configures signal handlers for the process.
 *
 * Configures the process to receive SIGALRM, SIGTERM, SIGHUP, SIGPIPE,
 * and SIGINT (Debug only).
 *
 * @return Zero on success, or a non-zero value on failure.
 *
 * @internal
 */
static int SetUpSignalHandlers(void)
{
   int result;
   struct sigaction sa;

   memset(&sa, 0, sizeof(struct sigaction));
   sa.sa_handler = SignalHandler;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags = 0; /* SA_ONESHOT; */

#ifdef HAVE_SA_RESTORER
   sa.sa_restorer = 0;
#endif

   result = sigaction(SIGALRM, &sa, 0);
   result |= sigaction(SIGTERM, &sa, 0);
   result |= sigaction(SIGPIPE, &sa, 0);

#ifdef DEBUG
   result |= sigaction(SIGINT, &sa, 0);
   result |= sigaction(SIGUSR1, &sa, 0);
#endif

   signal(SIGHUP, SignalHandler);
   /* result |= sigaction(SIGHUP,&sa,0); */

   return result;
}

/** Process main entry point.
 *
 * @param[in] argc - The number of command line arguments passed in @p argv.
 * @param[in] argv - An array of pointers to command line arguments.
 *
 * @return Zero on success, or a non-zero shell error code.
 *
 * @remarks This routine contains the main server loop.
 */
int main(int argc, char * argv[])
{
   SLPD_fdset fdset;
   int fdcount = 0;
   time_t curtime;
#if !HAVE_POLL
   struct timeval timeout;
#endif

#ifdef DEBUG
   xmalloc_init("/var/log/slpd_xmalloc.log", 0);
#endif

   SLPD_fdset_init(&fdset);

   /* Parse the command line */
   if (SLPDParseCommandLine(argc,argv))
      SLPDFatal("Invalid command line\n");

   /* make sure we are root */
   if (getuid() != 0)
      SLPDFatal("slpd must be started by root\n");

   /* make sure we are not already running */
   if (CheckPid(G_SlpdCommandLine.pidfile))
      SLPDFatal("slpd is already running. Check %s\n",
            G_SlpdCommandLine.pidfile);

   /* Initialize the preferences so we know if the log file is to be
      overwritten or appended.*/
   if (SLPDPropertyInit(G_SlpdCommandLine.cfgfile))
      SLPDFatal("slpd initialization failed during property load\n");

   /* make slpd run as a daemon */
   if (Daemonize(G_SlpdCommandLine.pidfile))
      SLPDFatal("Could not daemonize\n");

   /* initialize the log file */
   if (SLPDLogFileOpen(G_SlpdCommandLine.logfile, G_SlpdProperty.appendLog))
      SLPDFatal("Could not open logfile %s\n",G_SlpdCommandLine.logfile);

   /* seed the XID generator */
   SLPXidSeed();

   /* log startup message */
   SLPDLog("****************************************\n");
   SLPDLogTime();
   SLPDLog("SLPD daemon started\n");
   SLPDLog("****************************************\n");
   SLPDLog("Command line = %s\n", argv[0]);
   SLPDLog("Using configuration file = %s\n", G_SlpdCommandLine.cfgfile);
   SLPDLog("Using registration file = %s\n", G_SlpdCommandLine.regfile);
#ifdef ENABLE_SLPv2_SECURITY
   SLPDLog("Using SPI file = %s\n", G_SlpdCommandLine.spifile);
#endif

   /* initialize for the first time */
   SLPPropertyKeepFps();  /* do not close file descriptors */
   SLPDPropertyReinit();  /*So we get any property-related log messages*/
   if (
#ifdef ENABLE_SLPv2_SECURITY
         SLPDSpiInit(G_SlpdCommandLine.spifile) ||
#endif
         SLPDDatabaseInit(G_SlpdCommandLine.regfile)
         || SLPDIncomingInit()
         || SLPDOutgoingInit()
         || SLPDKnownDAInit())
      SLPDFatal("slpd initialization failed\n");
   SLPDLog("Agent Interfaces = %s\n", G_SlpdProperty.interfaces);
   if (G_SlpdProperty.port != SLP_RESERVED_PORT)
      SLPDLog("Using port %d instead of default %d\n", G_SlpdProperty.port, SLP_RESERVED_PORT);

   if (G_SlpdProperty.isDA && G_SlpdProperty.isDABackup)
      SLPDOpenDABackupFile();

   /* init watcher */
   SLPDDatabaseWatcher();

   /* tell systemd what we are ready */
   sd_notify(0, "READY=1");

   /* drop privileges to reduce security risk */
   if (DropPrivileges())
      SLPDFatal("Could not drop privileges\n");

   /* Setup signal handlers */
   if (SetUpSignalHandlers())
      SLPDFatal("Error setting up signal handlers.\n");

   if (G_SlpdProperty.isDA && G_SlpdProperty.isDABackup && DABackupfp)
   {
      SLPDDatabaseReadDABackup(DABackupfp);
   }

   if((G_SlpdProperty.isDA) && (G_SlpdProperty.DASyncReg))
   {
      /* HACK: at that point in time all outgoing sockets are DA connections
       * and the incoming sockets are our interfaces */
      SLPDLog("Pulling service list from other DAs...\n");
      SLPDSocket* sock = (SLPDSocket*)G_OutgoingSocketList.head;
      while (sock)
      {
         SLPDSocket* isock = (SLPDSocket*)G_IncomingSocketList.head;

         /* make sure we're not connecting to ourself */
         while (isock)
         {
             if (SLPNetCompareAddrs(&sock->peeraddr, &isock->peeraddr) == 0)
                break;
             isock = (SLPDSocket*)isock->listitem.next;
         }
         if (!isock)
	 {
	    int s = SLPNetworkConnectStream(&sock->peeraddr, 0);
	    if (s >= 0)
	    {
	       int result = getSLPServiceURLs(s, &sock->peeraddr, &sock->localaddr);
	       close(s);
	       if (result == 0)
		  break;
	    }
	 }
         sock = (SLPDSocket*)sock->listitem.next;
      }
   }

   /* Set up alarm to age database -- a shorter start, so SAs register with us quickly on our startup */
   alarm(2);

   /* Main loop */
   SLPDLog("Startup complete entering main run loop ...\n\n");
   G_SIGALRM   = 0;
   G_SIGTERM   = 0;
   G_SIGHUP    = 0;
#ifdef DEBUG
   G_SIGINT    = 0;
   G_SIGUSR1   = 0;
#endif

   while (G_SIGTERM == 0)
   {
      /* load the fdsets up with all valid sockets in the list  */
      SLPD_fdset_reset(&fdset);
      LoadFdSets(&G_IncomingSocketList, &fdset);
      LoadFdSets(&G_OutgoingSocketList, &fdset);

      /* before select(), check to see if we got a signal */
      if (G_SIGALRM || G_SIGHUP)
         goto HANDLE_SIGNAL;

      /* main select -- we time out every second so the outgoing retries can occur*/
      time(&curtime);
#if HAVE_POLL
      fdcount = poll(fdset.fds, fdset.used, 1000);
#else
      timeout.tv_sec = 1;
      timeout.tv_usec = 0;
      fdcount = select(fdset.highfd + 1, &fdset.readfds, &fdset.writefds, 0, &timeout);
#endif
      if (fdcount > 0) /* fdcount will be < 0 when interrupted by a signal */
      {
         SLPDIncomingHandler(&fdcount, &fdset);
         SLPDOutgoingHandler(&fdcount, &fdset);
         SLPDOutgoingRetry(time(0) - curtime);
      }
      else if (fdcount == 0)
         SLPDOutgoingRetry(time(0) - curtime);


HANDLE_SIGNAL:

      if (G_SIGHUP)
      {
         HandleSigHup();
         G_SIGHUP = 0;
      }
      if (G_SIGALRM)
      {
         HandleSigAlrm();
         G_SIGALRM = 0;
         alarm(SLPD_AGE_INTERVAL);
      }

#ifdef DEBUG
      if (G_SIGINT)
      {
         HandleSigInt();
         G_SIGINT = 0;
      }

      if (G_SIGUSR1)
      {
         HandleSigUsr1();
         G_SIGUSR1 = 0;
      }
#endif

      if (G_SlpdProperty.isDA && G_SlpdProperty.isDABackup && DABackupfp)
      {
         static time_t lastbck;
         time_t now;

         now = time(NULL);
         if (!lastbck)
            lastbck = now;
         if (now - lastbck > G_SlpdProperty.DABackupInterval)
         {
            SLPDLog("Updating registration backup file\n");
            SLPDDatabaseWriteDABackup(DABackupfp);
            lastbck = now;
         }
      }

   } /* End of main loop */

   /* Got SIGTERM */
   HandleSigTerm();

   return 0;
}
#endif   /* _WIN32 */

/*=========================================================================*/
